from __future__ import print_function
from builtins import input
import argparse
import os
import subprocess
import github
import yaml
from smush import GitLogStyleChecker


def arg_parser():
    """Return ArgumentParser for this application."""
    parser = argparse.ArgumentParser(
        description='Automate merging of feature branches.')

    parser.add_argument('topic_branch', nargs='?', metavar='topic_branch',
                        type=str, help='feature branch to merge')
    parser.add_argument('-n', '--new', action='store_true', default=False)
    parser.add_argument('-c', '--check', action='store_true', default=False)
    parser.add_argument('-d', '--delete-local', action='store_true', default=False)
    parser.add_argument('-p', '--profile')

    # Arguments that can be used instead of, or to override, configuration
    parser.add_argument('-b', '--base-branch', nargs='?', metavar='base branch', type=str)
    parser.add_argument('-o', '--github-owner', nargs='?', metavar='Github owner', type=str)
    parser.add_argument('-r', '--github-repo', nargs='?', metavar='Github repo', type=str)

    parser.add_argument('-v', '--version', action='store_true', help='display version and exit')

    parser.add_argument('--skip-style-check', action='store_true')
    parser.add_argument('--skip-pr-check', action='store_true')

    return parser


def load_config(profile=None, base_branch=None):
    """Return configuration, from YAML file, for this application.

    Args:
        profile (str, optional): name for profile (allows multiple configurations).
        base_branch (str, optional): base branch (supplied outside of configuration).

    Raises:
        Exception: If not able to read or parse the configuration file for any
                   reason or if "base branch" isn't set in the configuration
                   file.

    Returns:
        dict: Configuration information.
    """
    # Assemble configuration file name
    config_filename = '.smush'
    if profile:
        config_filename += '-' + profile
    config_filename += '.yml'

    # Attempt to load configuration file from user's home directory
    config_path = os.path.join(os.path.expanduser('~'), config_filename)
    config = {}

    # Make sure config file exists if needed
    if not os.path.isfile(config_path) and base_branch is None:
        raise Exception('Unable to open ~/{}: does it exist?'.format(config_filename))
    else:
        config = yaml.safe_load(open(config_path))

    # Verify base branch has been set in the config file
    if 'base branch' not in config and base_branch is None:
        raise Exception('Please set "base branch" in {} or specify via --base-branch.'.format(config_filename))

    return config


def check_topic_branch_commits(merge, skip_style_check=False, syntax_check_scripts=None):
    """Check topic branch commit(s) for issues.

    Args:
        merge (TopicMerge): Abstraction to make it easier to work with topic branch.
        skip_style_check (bool, optional): Whether to skip style check.

    Raises:
        Exception if not unmerged commits are found.
    """
    starting_branch = merge.active_branch()

    # Determine number of new commits and offer to rebase if greater than one
    if merge.unmerged_total() > 1:
        display_unmerged_commits(merge)

        print('There is more than one new commit present in dev branch.')
        if get_confirmation('Would you like to interactively rebase?'):
            merge.git.checkout(merge.topic_branch)
            subprocess.call(['git', 'rebase', '-i', 'HEAD~' + str(merge.unmerged_total())])
            print('Updating dev branch...')
            merge.git.push('--force')
            merge.git.checkout(starting_branch)
    elif merge.unmerged_total() == 0:
        raise Exception('No unmerged commits found.')

    # Display unmerged commits in dev branch
    display_unmerged_commits(merge)

    # Show commit style errors
    if not skip_style_check:
        checker = GitLogStyleChecker(merge.unmerged_log())
        print(checker.summarize_style_errors())

    # Show files with other errors
    if syntax_check_scripts:
        print('Checking unmerged files...')

        error_messages = []

        for filepath in merge.unmerged_files():
            filename, file_extension = os.path.splitext(filepath)
            file_extension = file_extension[1:]

            if isinstance(syntax_check_scripts, dict) and file_extension in syntax_check_scripts:
                print('Checking ' + filepath + '...')

                check_command = syntax_check_scripts[file_extension].format(filepath)
                process = subprocess.Popen(check_command, shell=True, stdout=subprocess.PIPE)
                process.wait()

                if process.returncode > 0:
                    error_messages.append('Error found in ' + filepath)

        if len(error_messages):
            for error_message in error_messages:
                print(error_message)

        print('Check complete.')


def check_for_pull_request(config, topic_branch):
    """ Check to make sure pull request for topic branch exists on Github.

    Uses Github owner/repository information and base branch from the
    application configuration to check Github to make sure an open pull
    request exists for the topic branch against the appropriate base branch.

    Args:
        config (dict): Application configuration.
        topic_branch (str): Topic branch.

    Raises:
        Exception: if the Github repository can't be found, a pull request
                   can't be found, or the pull request's base branch isn't
                   correct.
    """
    # Prepare to access Github repository
    try:
        repo = github.Github().get_repo("{}/{}".format(config['github owner'], config['github repo']))
    except github.UnknownObjectException:
        raise Exception('Github repository not found.')

    # Cycle through pulls request to find the first one for the topic branch
    pr = None
    for pull in repo.get_pulls(state='open'):
        if pull.head.ref == topic_branch:
            pr = pull

    # Report an error finding an appropriate pull request
    if not pr:
        raise Exception('Could not find pull request for this topic branch. Use --skip-pr-check option to skip.')
    elif pr.base.ref != config['base branch']:
        error_message = "The pull request's base is {}, not {}.".format(pr.base.ref, config['base branch'])
        raise Exception(error_message)


def display_unmerged_commits(merge):
    """Print unmerged commits with a blank line above and below."""
    print()
    print(merge.unmerged_log())
    print()


def get_confirmation(prompt):
    """Ask use to confirm an action."""
    confirm = input('{} [y/n] '.format(prompt))
    return len(confirm) and confirm[0].lower() == 'y'



class GitLogStyleChecker:
    def __init__(self, git_log_output):
        """Create git log style checker instance.

        Args:
            git_log_output (str): Git log (no flags set) output to parse/check.
        """
        self.commits = self.parse_and_check_git_log_output(git_log_output)

    def parse_and_check_git_log_output(self, git_log_output):
        """Return array of commit data and related errors.

        Parse git log output and return array of data relating to commits and
        any style errors found.

        Args:
            git_log_output (str): Git log output to parse and check.

        Returns:
            array: Commit information.
        """
        commits = []
        commit = {}
        blank_lines = 0

        for line in git_log_output.split("\n"):
            if line.startswith("commit "):
                # Note errors, append commit, reset bookkeeping variables
                if commit != {}:
                    self.check_commit_and_note_style_errors(commit)
                    commits.append(commit)
                    commit = {}
                    blank_lines = 0
                commit["hash"] = line.replace("commit ", "")
            elif line.startswith("Author: "):
                commit["author"] = line.replace("Author: ", "")
            elif line.startswith("Date:   "):
                commit["date"] = line.replace("Date:   ", "")
            elif "body" not in commit and line.strip() != "":
                commit["header"] = line[4:]
                commit["body"] = []
            elif line.strip() != "":
                # Note if body line appears too early
                if blank_lines == 1:
                    commit["no blank line before body"] = True
                commit["body"].append(line[4:])
            else:
                blank_lines += 1

        if commit != {}:
            self.check_commit_and_note_style_errors(commit)
            commits.append(commit)

        return commits

    def check_commit_and_note_style_errors(self, commit):
        """Add style error information to dict containing commit data.

        Error information, if found, is put into the "errors" key.

        Args:
            commit (dict): Commit data containing commit header and body.
        """
        errors = []

        # Check header
        if len(commit["header"]) > 50:
            errors.append("Subject line exceeds 50 characters.")

        # Check if body occurs directly after header
        if "no blank line before body" in commit:
            errors.append("If multi-line, needs blank line after subject line.")

        # Check lines
        line_number = 1
        for line in commit["body"]:
            if len(line) > 72:
                line_description = 'starting with "{}"'.format(line[:15])
                errors.append("Body line {} ({}) exceeds 72 characters.".format(line_number, line_description))
            line_number += 1

        # Add any errors to commit and append
        if len(errors):
            commit["errors"] = errors

    def summarize_style_errors(self):
        """Return text summary of errors found in commits parsed from the log.

        Returns:
            str: Text summary.
        """
        summary = ''

        # Create text summary of style errors
        for commit in self.commits:
            if "errors" in commit and len(commit["errors"]):
                summary += "Style error(s) found in commit {}:\n\n".format(commit["hash"])

                for line in commit["errors"]:
                    summary += "    {}\n".format(line)

                summary += "\n"

        if summary == '':
            summary += "No style errors found.\n"

        return summary



from __future__ import print_function
from builtins import super
import os
import git


class CurrentRepo(object):
    def __init__(self, base_branch):
        """Create current repo helper instance.

        Args:
            base_branch (str): Base branch.
        """
        self.base_branch = base_branch

        try:
            # Throws error if not in a Git repository
            self.repo = git.Repo(path='.', search_parent_directories=True)
            self.git = git.cmd.Git('.')
        except git.exc.InvalidGitRepositoryError:
            raise Exception('Not a valid git repository.')

        if not self.local_branch_exists(base_branch):
            raise Exception("Base branch '{}' not checked out locally.".format(base_branch))

    def update_base_branch(self):
        """Update base branch and rebase topic branch."""
        # Make sure base branch is up to date
        print("Checking out base branch '{}'...".format(self.base_branch))
        self.git.checkout(self.base_branch)
        print('Updating base branch...')
        self.git.pull('--rebase')

    def create_topic_branch(self, topic_branch_name):
        """Create topic branch locally and remotely."""
        print("Creating topic branch locally...")
        self.git.checkout(self.base_branch)
        self.git.checkout('-b', topic_branch_name)
        print("Pushing topic branch to base branch's remote...")
        self.git.push('-u', self.base_branch_remote(), topic_branch_name)

    def active_branch(self):
        """Return name of active branch."""
        return self.repo.active_branch.name

    def local_branch_exists(self, branch):
        """Check whether a branch exists locally in the current repository.

        Args:
            branch (str): Name of branch that may or may not exist.

        Returns:
            bool: True if branch exists, otherwise false.
        """
        return branch in self.repo.branches

    def base_branch_remote(self):
        """Return remote of base branch."""
        return self.git.config('--get', 'branch.{}.remote'.format(self.base_branch))


class TopicMerge(CurrentRepo):
    def __init__(self, base_branch, topic_branch=None, delete_local=False):
        """Create topic branch merge helper instance.

        Args:
            base_branch (str): Base branch.
            topic_branch (str, optional): Topic branch.
            delete_local (bool, optional): Whether or not to delete topic
                branch after merge.
        """
        super(TopicMerge, self).__init__(base_branch)

        self.topic_branch = topic_branch
        self.delete_local = delete_local

        if not topic_branch:
            self.topic_branch = self.active_branch()
            print("Using active branch '{}' for topic branch.".format(self.topic_branch))

        if self.topic_branch == self.base_branch:
            raise Exception("Topic branch and base branch shouldn't be the same.")

    def rebase_topic_branch_and_push(self):
        """Rebase topic branch with work from base branch and push."""
        # Rebase topic branch
        print('Checking out topic branch..')
        self.git.checkout(self.topic_branch)
        print('Updating topic branch with work from base branch...')
        self.git.rebase(self.base_branch)

        # Push rebased version (so it'll get marked as merged later if on
        # Github)
        print('Pushing updated topic branch...')
        self.git.push('--force')

    def merge_and_cleanup(self):
        """Merge topic branch then delete remotely and, optionally, locally."""
        print('Checking out base branch and merging topic branch...')
        self.git.checkout(self.base_branch)
        self.git.merge('--ff-only', self.topic_branch)

        # Push merge and delete topic branch
        print('Pushing base branch with topic branch merged...')
        self.git.push()
        print('Deleting remote topic branch...')
        self.git.push('origin', ':{}'.format(self.topic_branch))

        # Optionally delete local topic branch
        if self.delete_local:
            print('Deleting local topic branch...')
            self.git.branch('-D', self.topic_branch)

    def check_out_topic_branch_from_remote(self):
        """Check out local version of topic branch."""
        self.git.checkout('-b', self.topic_branch, '{}/{}'.format(self.base_branch_remote(), self.topic_branch))

    def remote_branch_exists(self, branch):
        """Check whether a branch exists remotely using base branch's origin.

        Args:
            branch (str): Name of branch that may or may not exist.

        Returns:
            bool: True if branch exists, otherwise false.
        """
        try:
            self.git.show_ref("refs/remotes/{}/{}".format(self.base_branch_remote(), branch))
            return True
        except git.exc.GitCommandError:
            return False

    def unmerged_commits(self):
        return self.repo.iter_commits('{}/{}..{}'.format(self.base_branch_remote(), self.base_branch, self.topic_branch))

    def unmerged_log(self):
        """Return Git log output for unmerged commits."""
        return self.git.log('{}..{}'.format(self.base_branch, self.topic_branch))

    def unmerged_total(self):
        """Return number of unmerged commits."""
        return int(self.git.rev_list('--count', '{}..{}'.format(self.base_branch, self.topic_branch)))

    def unmerged_files(self):
        filepaths = []
        file_root = self.repo.git.rev_parse('--show-toplevel')

        for commit in self.unmerged_commits():
            for filename in commit.stats.files:
                filepath = os.path.join(file_root, filename)

                if os.path.isfile(filepath) and filepath not in filepaths:
                    filepaths.append(filepath)

        return filepaths



import os
import sys

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))

from smush.git_log_style_checker import GitLogStyleChecker
from smush.topicmerge import CurrentRepo, TopicMerge

__version__ = "0.0.6"



import os
from setuptools import setup

# Get requirements
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# Get text of the README file
with open('README.rst') as f:
    README = f.read()


setup(
  name='smush',
  packages=['smush'],
  version='0.0.6',
  license='MIT',
  description='Tool to automate merging of topic branches',
  author='Mike Cantelon',
  author_email='mcantelon@gmail.com',

  long_description=README,
  long_description_content_type='text/x-rst',

  url='https://github.com/artefactual-labs/smush',

  keywords=['git'],

  install_requires=requirements,

  scripts=['bin/smush'],
)
